# cody.py (Explorer Workshop)
# Copyright: 2024 Bradan Lane STUDIO
#
# Open Source License: MIT

'''
This CircuitPython program is a sample from the 'Hack the Badge' workshop

Implementation Notes
--------------------
**Hardware:**
    * one of the Explorer Badge designs

**Software and Dependencies:**
    * CircuitPython firmware (UF2): https://circuitpython.org
    * Adafruit CircuitPython Libraries: https://circuitpython.org/libraries

**Open Source Repository:**
    * GitLab: https://gitlab.com/bradanlane/exp_tutorial
    
'''

import gc
import json
import time

import board
import time
import digitalio
import audiopwmio
import audiobusio
import synthio
import ulab.numpy as np # provides lots of numeric data processing

USE_I2S_AUDIO = False


#data_size = gc.mem_free()
with open('data/midi.json', 'r') as file:
    midi_data = json.load(file)
#data_size = data_size - gc.mem_free()
#print(f"JSON data size is {data_size:d}")
#time.sleep(2)

BPM = 105
DURATION64 = int(((1000 * 60) / BPM) / (64/4))
print(f"Duration of a 1/64th notes is {DURATION64:d} ms")



# create a basic synth
SAMPLE_SIZE = 1024
HALF_PERIOD = (SAMPLE_SIZE // 2)

synth = synthio.Synthesizer(sample_rate=22050)

env_lead = synthio.Envelope(attack_time=0.05, decay_time=0.1, release_time=0.1, attack_level=1.0, sustain_level=0.1)
env_base = synthio.Envelope(attack_time=0.025, decay_time=0.2, release_time=0.05, attack_level=1.0, sustain_level=0.1)
envelopes = [env_lead, env_base]

if USE_I2S_AUDIO:
    audio = audiobusio.I2SOut(board.I2S_BCK, board.I2S_LRCK, board.I2S_DATA)
    
    SAMPLE_VOLUME_LEAD = int(32767 * 0.1)
    SAMPLE_VOLUME_BASE = int(32767 * 0.05)
    
    wave_sine = np.array(np.sin(np.linspace(0, 2*np.pi, SAMPLE_SIZE, endpoint=False)) * SAMPLE_VOLUME_LEAD, dtype=np.int16)
    wave_saw = np.linspace(SAMPLE_VOLUME_BASE, -SAMPLE_VOLUME_BASE, num=SAMPLE_SIZE, dtype=np.int16)
    wave_tri = np.concatenate((np.linspace(-SAMPLE_VOLUME_BASE, SAMPLE_VOLUME_BASE, num=HALF_PERIOD, dtype=np.int16), np.linspace(SAMPLE_VOLUME_BASE, -SAMPLE_VOLUME_BASE, num=HALF_PERIOD, dtype=np.int16)))
    wave_square = np.concatenate((np.full(HALF_PERIOD, SAMPLE_VOLUME_BASE, dtype=np.int16), np.full(HALF_PERIOD, -SAMPLE_VOLUME_BASE, dtype=np.int16)))
    waves = [wave_sine, wave_saw]
    sound_enable = None
else:
    audio = audiopwmio.PWMAudioOut(board.SPEAKER)
    
    SAMPLE_VOLUME = 32767 # 100% as a 16bit integer
    
    #wave_sine = np.array(np.sin(np.linspace(0, 2*np.pi, SAMPLE_SIZE, endpoint=False)) * SAMPLE_VOLUME, dtype=np.int16)
    #wave_saw = np.linspace(SAMPLE_VOLUME, -SAMPLE_VOLUME, num=SAMPLE_SIZE, dtype=np.int16)
    #wave_tri = np.concatenate((np.linspace(-SAMPLE_VOLUME, SAMPLE_VOLUME, num=HALF_PERIOD, dtype=np.int16), np.linspace(SAMPLE_VOLUME, -SAMPLE_VOLUME, num=HALF_PERIOD, dtype=np.int16)))
    wave_square = np.concatenate((np.full(HALF_PERIOD, SAMPLE_VOLUME, dtype=np.int16), np.full(HALF_PERIOD, -SAMPLE_VOLUME, dtype=np.int16)))
    waves = [wave_square, wave_square]

    sound_enable = digitalio.DigitalInOut(board.SPEAKER_EN)
    sound_enable.direction = digitalio.Direction.OUTPUT
    sound_enable.value = True # we start with the amplifier OFF
# end AUDIO setup

audio.play(synth)


# because we can have multiple notes at once, we need to track them
active = {} # dictionary used to track notes

def note_on(note:int=60, channel:int=0):
    global waves, envelopes, synth, sound_enable, active
    frequency = synthio.midi_to_hz(note)
    synth_note = synthio.Note(frequency, waveform=waves[channel], envelope=envelopes[channel])
    synth.press(synth_note)
    # save the note so we can match it up when we stop playing it
    active[frequency] = synth_note

def note_off(note:int=60, channel:int=0):
    global synth, sound_enable, active
    frequency = synthio.midi_to_hz(note)
    # remove an active note (frequency) from the polyphonic output
    synth_note = active.get(frequency, None)
    if synth_note is not None:
        synth.release(synth_note)
        active.pop(frequency)

def note_off_all():
    global synth, sound_enable, active
    # remove all existing notes from the polyphonic output
    for entry in active:
        synth_note = active.get(entry, None)
        synth.release(synth_note)
        active.pop(entry)



timer = 0

try:
    for i in range(2):
        
        for event in midi_data:
            # event = [timecode, event, data1, data2]
            # if the new event has not yet happened, wait until it is time
            if event[0] > timer:
                #print(f"waiting {((event[0] - timer) * DURATION64):d} milliseconds")
                time.sleep(((event[0] - timer) * DURATION64) / 1000)
                #print(f"Time Index {event[0]:8d} ms")
            timer = event[0]
            
            # process event
            note = event[2] + 12 # 1 octave shift
            channel = event[1] & 0xF
            if ((event[1] & 0x90) == 0x90):	# note on
                #print(f"NOTE ON:  {(event[1]-0x90):d} {event[2]:2d} {event[3]:03d}")
                if (channel == 0) and ((timer % 256) == 0):
                    print()	# newline for each 4 bars
                print(channel, end="")
                note_on(note, channel)
            elif ((event[1] & 0x80) == 0x80):	# note off
                #print(f"NOTE OFF: {(event[1]-0x80):d} {event[2]:2d} {event[3]:03d}")
                note_off(note, channel)
            else:
                print(f"UNKNOWN:  {event[1]:3d} {event[2]:2d} {event[3]:03d}")

        note_off_all()
        timer = 0
        
        
finally:
    if sound_enable:
        sound_enable.value = False
    audio.deinit()
