# cody.py (Explorer Workshop)
# Copyright: 2024 Bradan Lane STUDIO
#
# Open Source License: MIT

'''
This CircuitPython program is a demo of the badge hardware

Implementation Notes
--------------------
**Hardware:**
    * one of the //BADGE_NAME// designs

**Software and Dependencies:**
    * CircuitPython firmware (UF2): https://circuitpython.org
    * Adafruit CircuitPython Libraries: https://circuitpython.org/libraries
'''


import board
import time
import random

from lib.simpletimer import SimpleTimer
from lib.explorer import expLed, expNeopixels, expTouch, expTone, expSynth, expIR, expDisplay

DISPLAY_MESSAGES = ["SETEC\nASTRONOMY", "YOUR\nNAME\nHERE", "DEFCON\nNEXTGEN", "DC32\nENGAGE", "PRIVACY\nNOT FOR\nSALE", "FIX\nTHE\nPLANET", "1%\nof\n1%", "EXPLORER\nBADGE"]
EIGHT_BALL = ["It is\ncertain", "It is\ndecidedly\nso", "Without\na doubt", "Yes\ndefinitely", "You may\nrely\non it", "As I\nsee it\nyes", "Most\nlikely", "Outlook\ngood", "Yes", "Signs\npoint\nto yes", "Reply hazy\ntry again", "Ask\nagain\nlater", "Better\nnot tell\nyou now", "Cannot\npredict\nnow", "Concentrate\nand ask\nagain", "Don't\ncount\non it", "My reply\nis no", "My\nsources\nsay no", "Outlook\nnot\nso good", "Very\ndoubtful"]

def run_demo(message=None):
    try:
        SIMPLE_SOUNDS = False   # toggle between simple Tones and full Synth sounds
        if message:
            DISPLAY_MESSAGE = message
        else:
            pos = random.randint(0, len(DISPLAY_MESSAGES)-1)
            DISPLAY_MESSAGE = DISPLAY_MESSAGES[pos]

        '''
        Example 1:
        load the expLed module and use it to blink the on board LED
        '''
        led = expLed()

        # let's make the LED flash
        print("LED blinks progressively faster")
        for i in range(5, 0, -1):
            led.blink(0.1 * i)
            time.sleep(0.2)


        '''
        Example 2:
        load the expNeopixels module and use it to sequence a list of colors
        '''
        neopixels = expNeopixels()

        # lets quickly cycle all the neopixels through a range of colors
        
        colors = [0x111111, 0xe0d000, 0xe07f07, 0xff1700, 0xd7009f, 0xaf01ff, 0x031ff7, 0x00fafa, 0x16e700, 0xffffff]
        COLOR_COUNT = 9
        counter = 0
        color = 0
        print("cycling though the colors of the neopixels")
        while True:
            counter += 1
            if not (counter % expNeopixels.COUNT):
                color = (color + 1) % 7
                if not color:
                    break   # when the color variable goes back to zero, break out of our loop
            neopixels.set(counter % expNeopixels.COUNT, colors[color])
            time.sleep(0.033)
        neopixels.clear()


        '''
        Example 3:
        load the expTouch module and read the initialized thresholds
        '''
        touches = expTouch()
        touches.update()
        print("Touch sensors raw:        ", [touches.raw(i) for i in range(expTouch.COUNT)])
        print("Touch sensors thresholds: ", [touches.threshold(i) for i in range(expTouch.COUNT)])


        '''
        Example 4:
        load either expTone or expSynth module and play C-Major
        '''
        if SIMPLE_SOUNDS:
            sound = expTone()
            print("Tone C-Major notes")
        else:
            sound = expSynth(expSynth.TRIANGLE)
            print("Synth C-Major chord")

        notes:list = [523.25, 587.33, 659.33, 698.46, 783.99, 880.00, 987.77, 1046.50, 1174.66]
        sound.note_on(notes[0])
        time.sleep(0.25)
        sound.note_on(notes[2])
        time.sleep(0.25)
        sound.note_on(notes[4])
        time.sleep(0.25)
        sound.note_on(notes[7])
        time.sleep(1.0)
        sound.note_off_all()



        '''
        Example 5:
        optionally load the expDisplay module and display some text on a RED background
        '''
        if DISPLAY_MESSAGE:
            count = 0
            parts = DISPLAY_MESSAGE.split("\n")
            for part in parts:
                count = count if count >= len(part) else len(part)
            # chose largest usable font
            vid = board.VID()
            if (vid == 2) or (vid == 4):
                BASE_NAME = "neuropolx"
            else:
                BASE_NAME = "graffiti"

            font_size = 18
            if   count > 10: font_size = 4
            elif count > 8:  font_size = 6
            elif count > 6:  font_size = 9
            elif count > 4:  font_size = 12

            fullname = f"{BASE_NAME}{font_size:d}.bdf"
            
            expDisplay.enable()
            display = expDisplay()
            print("Display '{}' on a red background".format(DISPLAY_MESSAGE))
            display.background(expDisplay.RED)
            display.text(fullname, DISPLAY_MESSAGE, expDisplay.WHITE, bgcolor=None, x=expDisplay.WIDTH / 2, y=expDisplay.HEIGHT / 2)
            wait = display.refresh()
            print("Display refresh will take {:f} seconds ".format(wait))
            time.sleep(wait + 0.5)
            while display.busy:
                print("Display is busy")
                time.sleep(0.5)
            time.sleep(0.5) # time for the ePaper to settle
            #display.disable()
        else:
            print("skipping the display")

        '''
        Example 6:
        scan the I2C bus and show what devices are attached
        '''
        time.sleep(1.0)
        # when using I2C, we need to lock the bus, before read/right tasks, and then unlock the bus
        have_accelerometer = False
        i2c_bus = board.I2C()
        try:    # wrapping code is 'try:except' lets us catch errors which would otherwise halt our code
            if i2c_bus.try_lock():
                print("Scanning I2C bus: ", end="") # the 'end' lets us suspend the default newline
                count = 0
                for x in i2c_bus.scan():
                    print(hex(x), end=" ")
                    count += 1
                    if x == 0x15:
                        have_accelerometer = True
                print("= {:d} device(s) found".format(count))
            i2c_bus.unlock()
        except Exception as e:
            print("We failed to scan the I2C bus - ", e)
        i2c_bus.unlock()

        # some badges do not have an accelerometer, so we will skip this section
        vid = board.VID()
        if ((vid == 2) or (vid == 4)) and have_accelerometer:
            from lib.explorer import expAccelerometer
            gyro = expAccelerometer()
            gyro.update()
            print ("Gyro: X:{:+05d} Y:{:+05d} Z:{:+05d} T:{:2d}".format(gyro.x, gyro.y, gyro.z, gyro.temperature))
        else:
            gyro = None

        ir = expIR()
        ir_message = [2, 3, 0, 255] # cmd, color, duration, delay (we are only interested in the color)
        
        print("ready")
        print("")

        msg_timer = SimpleTimer(10000)  # 10 second timer
        msg_timer.start()

        ir_timer = SimpleTimer(5000)	# 5 second timer
        ir_timer.start()
        
        counter = 0
        previous_counter = 0
        
        '''
            there is a conflict between pulseio used for IR and the touch sensing
            to solve this, we prevent IR receive if there are any active touches
            and then resume IR receive when there are not active touches
            (speculation is that multiple subsystems are using the RP2040 PIO)
        '''
        touch_count = 0

        while True:
            touches.update()
            if gyro:
                gyro.update()

            for i in range(expTouch.COUNT):
                if touches.touched(i):
                    print("touched  #{:d} = {:4d} on {:4d}".format(i+1, touches.raw(i), touches.threshold(i)))
                    neopixels.set(i, colors[i])
                    sound.note_on(notes[i])
                    if not touch_count:
                        ir.active = False
                    touch_count += 1
                if touches.released(i):
                    print("released #{:d} = {:4d} on {:4d}".format(i+1, touches.raw(i), touches.threshold(i)))
                    neopixels.set(i, 0x000000)
                    sound.note_off(notes[i])
                    touch_count -= 1
                    if not touch_count:
                        ir.active = True

            if gyro and gyro.interrupt:
                action = ("SHAKE" if gyro.shake else ("MOTION" if gyro.orientation else ""))
                print ("Gyro: X:{:+05d} Y:{:+05d} Z:{:+05d} T:{:2d} {}".format(gyro.x, gyro.y, gyro.z, gyro.temperature, action))
                # make noise
                for i in range(3,5):
                    sound.note_on(notes[i])
                time.sleep(1.0)
                for i in range(3,5):
                    sound.note_off(notes[i])

            if (not (ir is None)) and ir.available:
                code = ir.read()
                if code:
                    print("IR received:", code)
                    if len(code) >= 2:
                        color = colors[code[1] % COLOR_COUNT]
                        neopixels.fill(color)

            time.sleep(0.05)
            counter += 1
           
            if msg_timer.expired:
                # if we do the maths, we can determine how many loops we perform every 10 seconds (remembering we are sleeping for 0.05 seconds each loop)
                if gyro:
                    print ("Gyro: X:{:+05d} Y:{:+05d} Z:{:+05d} T:{:2d}".format(gyro.x, gyro.y, gyro.z, gyro.temperature))
                #print("Timer: loop #{:d} rate={:5.4}".format(counter, (10 / (counter - previous_counter))))
                #print(counter)
                previous_counter = counter
                msg_timer.start()
                
            if ir_timer.expired:
                if not ir is None:
                    color = random.randint(0, 7)
                    ir_message[1] = color
                    ir.write(ir_message)
                    print("IR transmitting:", ir_message)
                else:
                    print("IR skipped")
                ir_timer.start()
                
                
    finally:
        print("Disabling the ePaper display.")
        expDisplay.disable()	# prevent the display from rendering REPL output
        print("Good Bye")
    # end run_demo()


run_demo()  # this only gets executed if the program is not run as a module
#run_demo("Explorer\nBadge")  # this only gets executed if the program is not run as a module
