# cody.py (Explorer Workshop)
# Copyright: 2024 Bradan Lane STUDIO
#
# Open Source License: MIT

'''
This CircuitPython program turns the badge into a Magic 8-Ball

Used CircuitPython deep sleep to power off most of the badge.
Used the accelerometer to wake from sleep when shake is detected
Displays one of the random 8-Ball responses on the ePaper display
Goes into deep sleep

The demo only makes sense if the badge is battery powered.

IMPORTANT: the Explorer Badge has NeoPixels that are not turned off
by default. On the back of supported boards, there is instructions
for cutting a jumper. Cutting this jumper ties the power for NeoPixels
to the DISPLAY_EN pin. This allows the NeoPixels to be powered off
when the display is disabled. This is important for battery life.

NOTE: Not all Explorer Badge designs have the same features.

Implementation Notes
--------------------
**Hardware:**
    * one of the //BADGE_NAME// designs

**Software and Dependencies:**
    * CircuitPython firmware (UF2): https://circuitpython.org
    * Adafruit CircuitPython Libraries: https://circuitpython.org/libraries
'''


import board
import time
import random
import alarm

from lib.explorer import expAccelerometer, expDisplay

# there are two implementations of the Explorer Badge
DCNEXTGEN = 1
ECHALLENGECOIN = 2


EIGHT_BALL = ["It is\nCertain", 
              "It is\nDecidedly\nSo", 
              "Without\na Doubt", 
              "Yes\nDefinitely", 
              "Yes", 
              "You May\nRely\nOn It", 

              "As I\nSee It\nYes", 
              "Most\nLikely", 
              "Outlook\nGood", 
              "Signs\nPoint\nto Yes", 
              
              "Reply\nHazy\nTry\nAgain", 
              "Ask\nAgain\nLater", 
              "Better\nNot Tell\nYou Now", 
              "Cannot\nPredict\nNow", 
              "Concentrate\nand Ask\nAgain", 
              
              "Don't\nCount\non It", 
              "My\nReply\nIs No", 
              "My\nSources\nSay No", 
              "Outlook\nNot so\nGood", 
              "Very\nDoubtful"]

def show_answer(display):
    pos = random.randint(0, len(EIGHT_BALL)-1)
    answer = EIGHT_BALL[pos]

    count = 0
    parts = answer.split("\n")
    for part in parts:
        count = count if count >= len(part) else len(part)
        print(f"checking {part:12s} max width now '{count}'")
    # chose largest usable font
    if board.VID() == DCNEXTGEN:
        BASE_NAME = "graffiti"
        font_size = 18
        if count > 9:   font_size = 6
        elif count > 6: font_size_ = 9
        elif count > 4: font_size = 12
    else:
        BASE_NAME = "neuropolx"
        font_size = 18
        if count > 10:  font_size = 6
        elif count > 7: font_size_ = 9
        elif count > 6: font_size = 12

    fullname = f"{BASE_NAME}{font_size:d}.bdf"
    
    expDisplay.enable() # be sure the display is enabled
    msg = answer.replace("\n", " | ")
    print(f"Display '{msg}' on a black or red background")
    display.background(expDisplay.RED)
    display.text(fullname, answer, expDisplay.WHITE, bgcolor=None, x=expDisplay.WIDTH / 2, y=expDisplay.HEIGHT / 2)
    display.refresh(wait=True)

def show_bitmap(display):
    print("Display bitmap")
    display.background('/images/magic8ball.bmp')
    display.refresh(wait=True)
    time.sleep(0.5)        # give ePaper time to settle (prevents fading)
    
def run_demo():
    try:
        gyro = expAccelerometer(reserve_interrupt=True)
        show = gyro.interrupt
        display = expDisplay()

        while True:
            show = True
            while show:
                show_answer(display)
                print("wait 10 seconds")
                time.sleep(10)          # leave the answer on for a short while
                show = gyro.interrupt
                
            show_bitmap(display)
            display.disable()

            if board.VID() == ECHALLENGECOIN:
                print("deep sleep")
                alarm_pin = alarm.pin.PinAlarm(pin=board.GP28, value=False, pull=True)
                alarm.exit_and_deep_sleep_until_alarms(alarm_pin)
                # when the alarm wakes, the whole program will reload as start from the beginning
            else:
                counter = 0
                show = False
                while not show:
                    time.sleep(1)
                    counter += 1
                    print("waiting for shake ... ", counter)
                    show = gyro.interrupt
    except Exception as e:
        print("Oops. Things are not quire right. ", e)
    finally:
        print("Good Bye")
    # end run_demo()

run_demo()  # run the demo# this only gets executed if the program is not run as a module
